// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const FlowSampleCollectorSetTable = "Flow_Sample_Collector_Set"

// FlowSampleCollectorSet defines an object in Flow_Sample_Collector_Set table
type FlowSampleCollectorSet struct {
	UUID        string            `ovsdb:"_uuid"`
	Bridge      string            `ovsdb:"bridge"`
	ExternalIDs map[string]string `ovsdb:"external_ids"`
	ID          int               `ovsdb:"id" validate:"min=0,max=4294967295"`
	IPFIX       *string           `ovsdb:"ipfix"`
}
