// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const ICSBGlobalTable = "IC_SB_Global"

// ICSBGlobal defines an object in IC_SB_Global table
type ICSBGlobal struct {
	UUID        string            `ovsdb:"_uuid"`
	Connections []string          `ovsdb:"connections"`
	ExternalIDs map[string]string `ovsdb:"external_ids"`
	Options     map[string]string `ovsdb:"options"`
	SSL         *string           `ovsdb:"ssl"`
}
