// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const IGMPGroupTable = "IGMP_Group"

// IGMPGroup defines an object in IGMP_Group table
type IGMPGroup struct {
	UUID     string   `ovsdb:"_uuid"`
	Address  string   `ovsdb:"address"`
	Chassis  *string  `ovsdb:"chassis"`
	Datapath *string  `ovsdb:"datapath"`
	Ports    []string `ovsdb:"ports"`
}
