// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const DHCPv6OptionsTable = "DHCPv6_Options"

type (
	DHCPv6OptionsType = string
)

var (
	DHCPv6OptionsTypeIpv6 DHCPv6OptionsType = "ipv6"
	DHCPv6OptionsTypeStr  DHCPv6OptionsType = "str"
	DHCPv6OptionsTypeMAC  DHCPv6OptionsType = "mac"
)

// DHCPv6Options defines an object in DHCPv6_Options table
type DHCPv6Options struct {
	UUID string            `ovsdb:"_uuid"`
	Code int               `ovsdb:"code" validate:"min=0,max=254"`
	Name string            `ovsdb:"name"`
	Type DHCPv6OptionsType `ovsdb:"type" validate:"oneof='ipv6' 'str' 'mac'"`
}
