// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const CTZoneTable = "CT_Zone"

// CTZone defines an object in CT_Zone table
type CTZone struct {
	UUID          string            `ovsdb:"_uuid"`
	ExternalIDs   map[string]string `ovsdb:"external_ids"`
	TimeoutPolicy *string           `ovsdb:"timeout_policy"`
}
