// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const MeterTable = "Meter"

type (
	MeterUnit = string
)

var (
	MeterUnitKbps  MeterUnit = "kbps"
	MeterUnitPktps MeterUnit = "pktps"
)

// Meter defines an object in Meter table
type Meter struct {
	UUID  string    `ovsdb:"_uuid"`
	Bands []string  `ovsdb:"bands" validate:"min=1"`
	Name  string    `ovsdb:"name"`
	Unit  MeterUnit `ovsdb:"unit" validate:"oneof='kbps' 'pktps'"`
}
