// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const StaticMACBindingTable = "Static_MAC_Binding"

// StaticMACBinding defines an object in Static_MAC_Binding table
type StaticMACBinding struct {
	UUID               string `ovsdb:"_uuid"`
	IP                 string `ovsdb:"ip"`
	LogicalPort        string `ovsdb:"logical_port"`
	MAC                string `ovsdb:"mac"`
	OverrideDynamicMAC bool   `ovsdb:"override_dynamic_mac"`
}
