// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const AvailabilityZoneTable = "Availability_Zone"

// AvailabilityZone defines an object in Availability_Zone table
type AvailabilityZone struct {
	UUID string `ovsdb:"_uuid"`
	Name string `ovsdb:"name"`
}
