// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const ChassisTable = "Chassis"

// Chassis defines an object in Chassis table
type Chassis struct {
	UUID                string            `ovsdb:"_uuid"`
	Encaps              []string          `ovsdb:"encaps" validate:"min=1"`
	ExternalIDs         map[string]string `ovsdb:"external_ids"`
	Hostname            string            `ovsdb:"hostname"`
	Name                string            `ovsdb:"name"`
	NbCfg               int               `ovsdb:"nb_cfg"`
	OtherConfig         map[string]string `ovsdb:"other_config"`
	TransportZones      []string          `ovsdb:"transport_zones"`
	VtepLogicalSwitches []string          `ovsdb:"vtep_logical_switches"`
}
