// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const AddressSetTable = "Address_Set"

// AddressSet defines an object in Address_Set table
type AddressSet struct {
	UUID        string            `ovsdb:"_uuid"`
	Addresses   []string          `ovsdb:"addresses"`
	ExternalIDs map[string]string `ovsdb:"external_ids"`
	Name        string            `ovsdb:"name"`
}
