// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const MulticastGroupTable = "Multicast_Group"

// MulticastGroup defines an object in Multicast_Group table
type MulticastGroup struct {
	UUID      string   `ovsdb:"_uuid"`
	Datapath  string   `ovsdb:"datapath"`
	Name      string   `ovsdb:"name"`
	Ports     []string `ovsdb:"ports"`
	TunnelKey int      `ovsdb:"tunnel_key" validate:"min=32768,max=65535"`
}
