// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const FDBTable = "FDB"

// FDB defines an object in FDB table
type FDB struct {
	UUID    string `ovsdb:"_uuid"`
	DpKey   int    `ovsdb:"dp_key" validate:"min=1,max=16777215"`
	MAC     string `ovsdb:"mac"`
	PortKey int    `ovsdb:"port_key" validate:"min=1,max=16777215"`
}
