// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const MeterBandTable = "Meter_Band"

type (
	MeterBandAction = string
)

var MeterBandActionDrop MeterBandAction = "drop"

// MeterBand defines an object in Meter_Band table
type MeterBand struct {
	UUID      string          `ovsdb:"_uuid"`
	Action    MeterBandAction `ovsdb:"action" validate:"oneof='drop'"`
	BurstSize int             `ovsdb:"burst_size" validate:"min=0,max=4294967295"`
	Rate      int             `ovsdb:"rate" validate:"min=1,max=4294967295"`
}
