// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const PortBindingTable = "Port_Binding"

// PortBinding defines an object in Port_Binding table
type PortBinding struct {
	UUID             string            `ovsdb:"_uuid"`
	Address          string            `ovsdb:"address"`
	AvailabilityZone string            `ovsdb:"availability_zone"`
	Encap            *string           `ovsdb:"encap"`
	ExternalIDs      map[string]string `ovsdb:"external_ids"`
	Gateway          string            `ovsdb:"gateway"`
	LogicalPort      string            `ovsdb:"logical_port"`
	TransitSwitch    string            `ovsdb:"transit_switch"`
	TunnelKey        int               `ovsdb:"tunnel_key" validate:"min=1,max=32767"`
}
