// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const MirrorTable = "Mirror"

type (
	MirrorFilter = string
	MirrorType   = string
)

var (
	MirrorFilterFromLport MirrorFilter = "from-lport"
	MirrorFilterToLport   MirrorFilter = "to-lport"
	MirrorTypeGre         MirrorType   = "gre"
	MirrorTypeErspan      MirrorType   = "erspan"
)

// Mirror defines an object in Mirror table
type Mirror struct {
	UUID        string            `ovsdb:"_uuid"`
	ExternalIDs map[string]string `ovsdb:"external_ids"`
	Filter      MirrorFilter      `ovsdb:"filter" validate:"oneof='from-lport' 'to-lport'"`
	Index       int               `ovsdb:"index"`
	Name        string            `ovsdb:"name"`
	Sink        string            `ovsdb:"sink"`
	Type        MirrorType        `ovsdb:"type" validate:"oneof='gre' 'erspan'"`
}
