// Code generated by "libovsdb.modelgen"
// DO NOT EDIT.

package ovsmodel

const ManagerTable = "Manager"

type (
	ManagerConnectionMode = string
)

var (
	ManagerConnectionModeInBand    ManagerConnectionMode = "in-band"
	ManagerConnectionModeOutOfBand ManagerConnectionMode = "out-of-band"
)

// Manager defines an object in Manager table
type Manager struct {
	UUID            string                 `ovsdb:"_uuid"`
	ConnectionMode  *ManagerConnectionMode `ovsdb:"connection_mode" validate:"omitempty,oneof='in-band' 'out-of-band'"`
	ExternalIDs     map[string]string      `ovsdb:"external_ids"`
	InactivityProbe *int                   `ovsdb:"inactivity_probe"`
	IsConnected     bool                   `ovsdb:"is_connected"`
	MaxBackoff      *int                   `ovsdb:"max_backoff" validate:"omitempty,min=1000"`
	OtherConfig     map[string]string      `ovsdb:"other_config"`
	Status          map[string]string      `ovsdb:"status"`
	Target          string                 `ovsdb:"target"`
}
